@@ -14,6 +14,44 @@ module Database.Sqlite  (
                          LogFunction,
                          SqliteStatus (..),
                          SqliteStatusVerb (..),
+    -- * Basic usage guide
+    -- |
+    --
+    -- Note that the example code shown here is a low level interface
+    -- usage. Let's create a small demo sqlite3 database which we will
+    -- use in our program:
+    --
+    -- > $ sqlite3 ~/test.db
+    -- > sqlite> create table t1(a,b);
+    -- > sqlite> insert into t1(a,b) values (1,1);
+    -- > sqlite> insert into t1(a,b) values (2,2);
+    -- > sqlite> select * from t1;
+    -- > 1|1
+    -- > 2|2
+    --
+    -- Now let's write code using the functions in this module to
+    -- fetch the rows from the table:
+    --
+    -- > {-#LANGUAGE OverloadedStrings#-}
+    -- >
+    -- > import Database.Sqlite
+    -- > import Data.Text
+    -- >
+    -- > main :: IO ()
+    -- > main = do
+    -- >   conn <- open "/home/sibi/test.db"
+    -- >   smt <- prepare conn "select * from t1;"
+    -- >   row1 <- step smt >> columns smt
+    -- >   row2 <- step smt >> columns smt
+    -- >   print (row1, row2)
+    -- >   finalize smt
+    -- >   close conn
+    --
+    -- On executing the above code:
+    --
+    -- > $ ./demo-program
+    -- > $ ([PersistInt64 1,PersistInt64 1],[PersistInt64 2,PersistInt64 2])
+
                          open,
                          close,
                          prepare,
@@ -35,7 +73,9 @@ module Database.Sqlite  (
                          freeLogFunction,
                          config,
                          status,
-                         softHeapLimit
+                         softHeapLimit,
+                         enableExtendedResultCodes,
+                         disableExtendedResultCodes
                         )
     where
 
@@ -191,6 +231,7 @@ sqlError maybeConnection functionName error = do
 
 foreign import ccall "sqlite3_open_v2"
   openC :: CString -> Ptr (Ptr ()) -> Int -> CString -> IO Int
+
 openError :: Text -> IO (Either Connection Error)
 openError path' = do
     let flag = sqliteFlagReadWrite .|. sqliteFlagCreate .|. sqliteFlagUri
@@ -228,6 +269,28 @@ close database = do
     ErrorOK -> return ()
     _ -> sqlError (Just database) "close" error
 
+foreign import ccall "sqlite3_extended_result_codes"
+  sqlite3_extended_result_codesC :: Ptr () -> Int -> IO Int
+
+
+-- @since 2.9.2
+enableExtendedResultCodes :: Connection -> IO ()
+enableExtendedResultCodes con@(Connection _ (Connection' database)) =  do
+  error <- sqlite3_extended_result_codesC database 1
+  let err = decodeError error
+  case err of
+    ErrorOK -> return ()
+    _ -> sqlError (Just con) "enableExtendedResultCodes" err
+
+-- @since 2.9.2
+disableExtendedResultCodes :: Connection -> IO ()
+disableExtendedResultCodes con@(Connection _ (Connection' database)) =  do
+  error <- sqlite3_extended_result_codesC database 0
+  let err = decodeError error
+  case err of
+    ErrorOK -> return ()
+    _ -> sqlError (Just con) "disableExtendedResultCodes" err
+
 foreign import ccall "sqlite3_prepare_v2"
   prepareC :: Ptr () -> CString -> Int -> Ptr (Ptr ()) -> Ptr (Ptr ()) -> IO Int
 prepareError :: Connection -> Text -> IO (Either Statement Error)
