@@ -6,4 +6,4 @@ import Control.Monad.Logger
 $(return []) -- just force TH to run
 
 main :: IO ()
-main = runStderrLoggingT $ withSqliteConn ":memory:" $ const $ return ()
+main = runStderrLoggingT $ withSqliteConn ":memory:" $ runSqlConn waitForDatabase
